## Migrating from Terrajet to Upjet

As [Terrajet] is being [deprecated][terrrajet-deprecation] in favor of [Upjet],
new Crossplane providers are expected to be generated using Upjet's [provider
template repository](https://github.com/upbound/upjet-provider-template), and
this guide outlines the steps needed to convert an existing Terrajet-based
Crossplane provider to an Upjet-based one. While this guide describes the
steps needed to update the `github.com/crossplane/terrajet` Go module
dependency with the `github.com/upbound/upjet` module, especially if you
have not customized your provider's CI pipelines and architecture, and
you want to utilize all the new features offered by Upjet 
(such as documentation/example manifest generation), 
you may also prefer to [bootstrap a new provider] from the
[Upjet template repository] and copy your existing resource configurations
under the `config` folder of your existing provider to the new provider.
The steps involved are similar in both approaches but this guide provides
more flexibility as we have indicated certain steps as optional to the
migration process itself.

### Overview of Terrajet-Upjet Migration Steps

1. Replace the `github.com/crossplane/terrajet` Go module dependency with the
latest release of `github.com/upbound/upjet`. You may run the following in the
provider repo's root (where the `go.mod` file lives)
```shell
go mod edit -droprequire github.com/crossplane/terrajet
go get github.com/upbound/upjet
```

2. Some code that will be generated by Upjet requires go1.18 or later. If your
provider's Go module declares go1.17 or smaller, you will need to update your
module's Go version, and run a `go mod tidy` to make sure that `go.mod` matches
the source code in the provider's module:
```shell
go mod edit -go=1.19
go mod tidy
```

3. Replace all occurrences of `crossplane/terrajet` with `upbound/upjet` in the
code base. A sample command (to be run from the repo root) could be as follows,
or you may use any other tool at your convenience:
```shell
find . -type f -not \( -path './.git/*' -or -path './.cache/*' -or -path './.work/*' \) -exec sed -i '' 's|crossplane/terrajet|upbound/upjet|g' {} \;  
```

4. Now, you may attempt to run Upjet's code generation pipelines:
```shell
make generate
```
At this step, you will most probably get some compile errors. Here are some
potential changes that you need to make, assuming the provider was
bootstrapped using Terrajet's now deprecated [provider template
repository](https://github.com/crossplane-contrib/provider-jet-template):
- In `config/provider.go`:
[pkg/config.NewProviderWithSchema](https://github.com/crossplane/terrajet/blob/8d0ed485f9511b65a8f3a83801092bcae60678dd/pkg/config/provider.go#L152)
has been removed from Upjet. You may instead use
[pkg/config.NewProvider](https://github.com/upbound/upjet/blob/c82119f5ef342f752406a0ed38264940b02e795f/pkg/config/provider.go#L172).
For an example invocation, please see
[here](https://github.com/upbound/upjet-provider-template/blob/d34119409586f6205ec8ed4b9b2c2481c74bf07e/config/provider.go#L29).
- In `config/provider.go`:
[pkg/config.WithDefaultResourceFn](https://github.com/crossplane/terrajet/blob/8d0ed485f9511b65a8f3a83801092bcae60678dd/pkg/config/provider.go#L143)
has been removed from Upjet. You may instead use
[pkg/config.WithDefaultResourceOptions](https://github.com/upbound/upjet/blob/c82119f5ef342f752406a0ed38264940b02e795f/pkg/config/provider.go#L154).
For an example invocation, please see
[here](https://github.com/upbound/upjet-provider-template/blob/d34119409586f6205ec8ed4b9b2c2481c74bf07e/config/provider.go#L31).
- In `config/provider.go`: If your provider does not have "base" API packages
(like the `ProviderConfig` packages) at both `v1alpha1` and `v1beta1` versions,
then you may need to override the provider's [`BasePackages`
configuration](https://github.com/upbound/upjet/blob/7e84c638a8bc5c93c6da3cf9420f961f165dd05d/pkg/config/provider.go#L76)
configuration using
[pkg/config.WithBasePackages](https://github.com/upbound/upjet/blob/7e84c638a8bc5c93c6da3cf9420f961f165dd05d/pkg/config/provider.go#L146).
This is because Upjet
[assumes](https://github.com/upbound/upjet/blob/7e84c638a8bc5c93c6da3cf9420f961f165dd05d/pkg/config/common.go#L18)
both `apis/v1alpha1` `apis/v1beta1` base API packages and the
`internal/controller/providerconfig` controller package exist for your provider.
If, for example, only the `apis/v1alpha1` base API package is available for your
provider, then you may configure your provider with:
```go
config.WithBasePackages(config.BasePackages{
			APIVersion: []string{
				// Default package for ProviderConfig APIs
				"apis/v1alpha1",
			},
			Controller: []string{
				// Default package for ProviderConfig controllers
				"internal/controller/providerconfig",
			},
		})
```
as a
[config.ProviderOption](https://github.com/upbound/upjet/blob/7e84c638a8bc5c93c6da3cf9420f961f165dd05d/pkg/config/provider.go#L115)
passed to
[pkg/config.NewProvider](https://github.com/upbound/upjet/blob/c82119f5ef342f752406a0ed38264940b02e795f/pkg/config/provider.go#L172).

5. When you try to build the provider package, you may also encounter some
compilation errors regarding managed resource configurations:
- In various `apis/<resource API group>/<API
version>/zz_generated.resolvers.go`: You may observe errors referring to
undeclared `spec` struct fields with names containing acronyms such as:
```shell
apis/ec2/v1alpha2/zz_generated.resolvers.go:195:38: mg.Spec.ForProvider.SecurityGroupIdRefs undefined (type InstanceParameters has no field or method SecurityGroupIdRefs)
```
You will also need to fix the [reference configurations] for such managed
resources using the [resource configuration API].

6. You will need to make changes in your `Makefile` to get your make targets
working again:
- You need to update the declared Go version with the `GO_REQUIRED_VERSION`
make variable.
- You need to update the declared Go linter version with
the `GOLANGCILINT_VERSION` make variable.
```shell
GO_REQUIRED_VERSION ?= 1.19
GOLANGCILINT_VERSION ?= 1.50.0
```

This make variables must be available when the make targets consuming them
are run, so please make sure to include them at the head of the file. For an
example, please check the Upjet's provider template's [Makefile].

7. Now, you may want to run `make lint` to reveal and resolve any linter issues.
You will probably need to run `gofmt -s -w` on certain Go sources to properly
format them.

After these steps, you should now be able to build your provider package with
`make build`. Please also keep in mind that you may need to update your repo's
CI pipelines especially if you've updated your provider module's Go version.

8. Now, you may want to build your provider package and test the fresh package
   with your provider's managed resources.

Having successfully migrated from the `github.com/crossplane/terrajet`
Go module to the `github.com/upbound/upjet` module, you may now consider
enabling some new and advanced features only available in Upjet:

9. You may consider enabling metadata extraction from the [Terraform registry].
In order to enable metadata extraction for the provider, you will need to run
Upjet's metadata scraper. This can be achieved by including a Go generate
comment in your provider's `apis/generate.go` with something similar to:
```go
//go:generate go run github.com/upbound/upjet/cmd/scraper -n ${TERRAFORM_PROVIDER_SOURCE} -r ../.work/${TERRAFORM_PROVIDER_SOURCE}/${TERRAFORM_DOCS_PATH} -o ../config/provider-metadata.yaml
```
This generate directive depends on some Makefile changes like adding the
`pull-docs` target and declaring certain Makefile variables. For details,
please refer to Upjet's provider template [Makefile](https://github.com/upbound/upjet-provider-template/blob/main/Makefile)
and [apis/generate.go](https://github.com/upbound/upjet-provider-template/blob/main/apis/generate.go).

Now, when you run a `make generate`, you should find the extracted metadata at
`config/provider-metadata.yaml`.

10. Now that your provider extracts Terraform registry metadata, you can use it
to generate CRD documentation and to generate example CR manifests. In order to
do this, you will need to configure your provider with the scraped metadata
as follows:
- In `config/provider.go`: You need to load the provider's metadata from the
`config/provider-metadata.yaml` file.
You can easily do so with the `go:embed` directive:
```go
//go:embed provider-metadata.yaml
var providerMetadata []byte
```
- In `config/provider.go`: You need to make the provider metadata available
to the provider configuration. This is done by passing the metadata bytes
(stored in the variable `providerMetadata` above) to the
`config.NewProvider` call. For an example, please refer to the template
repo's [config/provider.go](https://github.com/upbound/upjet-provider-template/blob/d34119409586f6205ec8ed4b9b2c2481c74bf07e/config/provider.go#L29).

Now, when you run `make generate`, the generated CRDs for your provider should
have documentation scraped from the Terraform registry. And you should
have some example manifests generated for your managed resources under
the path `examples-generated` in the repo root.

Please also take a look at the template repo's [apis/generate.go](https://github.com/upbound/upjet-provider-template/blob/main/apis/generate.go)
for some other `go:generate` directives for cleaning up these newly
generated artifacts.

11. You may also want to update your provider's `build` submodule to the latest version by running a:
```shell
git submodule update --remote --merge
```

12. If you would like to enable a shared gRPC server for your provider
(which implies that the Terraform CLI will not fork multiple native provider
processes to make requests and instead, share a common instance across
all requests to improve the performance of the Crossplane provider), or
if you would like to integrate the single image building process,
please adapt your provider package's Dockerfile
(`cluster/images/<provider name>/Dockerfile`) and
Makefile (`cluster/images/<provider name>/Makefile`)
to the template repo's [associated files](https://github.com/upbound/upjet-provider-template/tree/main/cluster/images/upjet-provider-template).
You will also need to adapt your provider's Makefile to match
the template repo's [Makefile](https://github.com/upbound/upjet-provider-template/blob/main/Makefile).


[Upjet]: https://github.com/upbound/upjet
[Terrajet]: https://github.com/crossplane/terrajet
[terrrajet-deprecation]: https://github.com/crossplane/terrajet/issues/308
[resource configuration API]:
    https://github.com/upbound/upjet/blob/7e84c638a8bc5c93c6da3cf9420f961f165dd05d/pkg/config/resource.go#L258
[reference configurations]: https://github.com/upbound/upjet/blob/c82119f5ef342f752406a0ed38264940b02e795f/pkg/config/resource.go#L293
[Upjet template repository]: https://github.com/upbound/upjet-provider-template
[Makefile]: https://github.com/upbound/upjet-provider-template/blob/d34119409586f6205ec8ed4b9b2c2481c74bf07e/Makefile#L41
[Terraform registry]: https://registry.terraform.io/
[bootstrap a new provider]: https://github.com/upbound/upjet/blob/main/docs/generating-a-provider.md
